# Copyright (c) 2014 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from u2fval.model import Client, User, Device
from u2fval.core.exc import (BadInputException, NoEligibleDevicesException,
                             DeviceCompromisedException)
from u2flib_server.u2f_v2 import (start_register, complete_register,
                                  start_authenticate, verify_authenticate)
from u2flib_server.utils import rand_bytes
from sqlalchemy.orm import exc
from datetime import datetime
import logging


__all__ = ['U2FController']
log = logging.getLogger(__name__)


class U2FController(object):

    def __init__(self, session, memstore, client_name, metadata,
                 require_trusted=True):
        self._session = session
        self._memstore = memstore
        try:
            self._client = session.query(Client) \
                .filter(Client.name == client_name).one()
        except exc.NoResultFound:
            raise KeyError('No Client with name %s found' % client_name)
        self._metadata = metadata
        self._require_trusted = require_trusted

    def _get_user(self, username):
        return self._session.query(User) \
            .filter(User.client_id == self._client.id) \
            .filter(User.name == username).first()

    def _get_device(self, handle):
        return self._session.query(Device).join(Device.user) \
            .filter(User.client_id == self._client.id) \
            .filter(Device.handle == handle).first()

    def _get_or_create_user(self, username):
        user = self._get_user(username)
        if user is None:
            user = User(username)
            self._client.users.append(user)
            log.info('User created: "%s/%s"', self._client.name, username)
        return user

    @property
    def client_name(self):
        return self._client.name

    def get_trusted_facets(self):
        return {
            'trustedFacets': [{
                'version': {'major': 1, 'minor': 0},
                'ids': self._client.valid_facets
            }]
        }

    def delete_user(self, username):
        user = self._get_user(username)
        if user is not None:
            self._session.delete(user)
            log.info('User deleted: "%s/%s"', self._client.name, username)

    def register_start(self, username):
        # RegisterRequest
        register_request = start_register(self._client.app_id)
        self._memstore.store(self._client.id, username,
                             register_request.challenge,
                             {'request': register_request})

        # SignRequest[]
        sign_requests = []
        descriptors = []
        user = self._get_user(username)
        if user is not None:
            for dev in user.devices.values():
                sign_requests.append(
                    start_authenticate(dev.bind_data, 'check-only'))
                descriptors.append(dev.get_descriptor(
                    self._metadata.get_metadata(dev)))

        # To support multiple versions, add more RegisterRequests.
        return [register_request], sign_requests, descriptors

    def register_complete(self, username, resp):
        memkey = resp.clientData.challenge
        data = self._memstore.retrieve(self._client.id, username, memkey)
        bind, cert = complete_register(data['request'], resp,
                                       self._client.valid_facets)
        attestation = self._metadata.get_attestation(cert)
        if self._require_trusted and not attestation.trusted:
            raise BadInputException('Device type is not trusted')
        user = self._get_or_create_user(username)
        dev = user.add_device(bind.json, cert, attestation.transports)
        log.info('User: "%s/%s" - Device registered: "%s"',
            self._client.name, username, dev.handle)
        return dev.handle

    def unregister(self, handle):
        dev = self._get_device(handle)
        self._session.delete(dev)
        log.info('User: "%s/%s" - Device unregistered: "%s"',
            self._client.name, dev.user.name, handle)

    def set_props(self, handle, props):
        dev = self._get_device(handle)
        dev.properties.update(props)

    def _do_get_descriptor(self, user, handle):
        if user is not None:
            dev = self._session.query(Device) \
                .filter(Device.user_id == user.id) \
                .filter(Device.handle == handle).first()
        if user is None or dev is None:
            raise BadInputException('No device matches descriptor: %s' % handle)
        return dev.get_descriptor(self._metadata.get_metadata(dev))

    def get_descriptor(self, username, handle):
        user = self._get_user(username)
        return self._do_get_descriptor(user, handle)

    def get_descriptors(self, username):
        user = self._get_user(username)
        if user is None:
            return []
        return [d.get_descriptor(self._metadata.get_metadata(d))
                for d in user.devices.values()]

    def authenticate_start(self, username, invalidate=False):
        user = self._get_user(username)
        if user is None or len(user.devices) == 0:
            log.info('User "%s" has no devices registered', username)
            raise NoEligibleDevicesException('No devices registered', [])

        sign_requests = []
        descriptors = []
        challenges = {}
        rand = rand_bytes(32)

        for handle, dev in user.devices.items():
            if not dev.compromised:
                challenge = start_authenticate(dev.bind_data, rand)
                sign_requests.append(challenge)
                descriptors.append(dev.get_descriptor(
                    self._metadata.get_metadata(dev)))
                challenges[handle] = {
                    'keyHandle': challenge.keyHandle,
                    'challenge': challenge
                }

        if not sign_requests:
            raise NoEligibleDevicesException(
                'All devices compromised',
                [d.get_descriptor() for d in user.devices.values()]
            )
        self._memstore.store(self._client.id, username, rand, challenges)
        return sign_requests, descriptors

    def authenticate_complete(self, username, resp):
        memkey = resp.clientData.challenge
        challenges = self._memstore.retrieve(self._client.id, username, memkey)
        user = self._get_user(username)
        for handle, data in challenges.items():
            if data['keyHandle'] == resp.keyHandle:
                dev = user.devices[handle]
                if dev.compromised:
                    raise BadInputException('Device is compromised')

                counter, presence = verify_authenticate(
                    dev.bind_data,
                    data['challenge'],
                    resp,
                    self._client.valid_facets
                )
                if presence == chr(0):
                    raise Exception('User presence byte not set!')
                if counter > (dev.counter or -1):
                    dev.counter = counter
                    dev.authenticated_at = datetime.now()
                    return handle
                dev.compromised = True
                raise DeviceCompromisedException('Device counter mismatch',
                                                 dev.get_descriptor())
        else:
            raise BadInputException('No device found for keyHandle: %s' %
                                    resp.keyHandle)



#    Copyright (C) 2014  Yubico AB
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function

import os
import sys
import argparse


def run_parser(parser):
    parser.add_argument('-i', '--interface', default='localhost',
                        help='network interface to bind to')
    parser.add_argument('-p', '--port', type=int, default=8080,
                        help='TCP port to bind to')
    parser.add_argument('-c', '--client', help='run in single client mode '
                        'using CLIENT')


def client_parser(parser):
    class NameFromFacetsAction(argparse.Action):
        def __call__(self, parser, args, values, option_string=None):
            if not values:
                if args.facets and len(args.facets) > 1:
                    values, args.facets = args.facets[-1], args.facets[:-1]
                else:
                    parser.error('argument %s is required' % self.dest)
            args.name = values

    client_subparsers = parser.add_subparsers(dest='action', help='subcommand')

    list_parser = client_subparsers.add_parser('list',
                                               help='list available clients')

    create_parser = client_subparsers.add_parser('create',
                                                 help='create a client')
    create_parser.add_argument('name', metavar='<name>', nargs='?',
                               action=NameFromFacetsAction,
                               help='the name of the client')
    create_parser.add_argument('-a', '--appId', required=True,
                               help='sets the appId')
    create_parser.add_argument('-f', '--facets', required=True, nargs='+',
                               metavar='FACET',
                               help='all valid facets for the client')

    show_parser = client_subparsers.add_parser('show',
                                               help='display a client')
    show_parser.add_argument('name', metavar='<name>',
                             help='the name of the client')

    update_parser = client_subparsers.add_parser('update',
                                                 help='update data for a '
                                                 'client')
    update_parser.add_argument('name', metavar='<name>', nargs='?',
                               action=NameFromFacetsAction,
                               help='the name of the client')
    update_parser.add_argument('-a', '--appId',
                               help='sets the appId')
    update_parser.add_argument('-f', '--facets', nargs='+', metavar='FACET',
                               help='all valid facets for the client')

    delete_parser = client_subparsers.add_parser('delete',
                                                 help='delete a client')
    delete_parser.add_argument('name', metavar='<name>',
                               help='the name of the client')


def db_parser(parser):
    parser.add_argument('action', choices=['init', 'upgrade'], help='subcommand')


def arg_parser():
    parser = argparse.ArgumentParser(
        description="Yubico U2F Validation Server",
        add_help=True
    )
    parser.add_argument('-c', '--config', help='specify an alternate '
                        'configuration file to use')
    parser.add_argument('-d', '--debug', action='store_true', dest='debug',
                        help='prints debug information to stdout')
    subparsers = parser.add_subparsers(dest='command',
                                       help='available commands')

    run_parser(subparsers.add_parser('run', help='run the server'))
    client_parser(subparsers.add_parser('client', help='manage clients'))
    db_parser(subparsers.add_parser('db', help='manage database'))

    return parser


def create_session(settings, echo=False):
    from sqlalchemy import create_engine
    from sqlalchemy.orm import sessionmaker
    engine = create_engine(settings['db'], echo=echo)

    Session = sessionmaker(bind=engine)
    return Session()


def handle_client(settings, args):
    from u2fval.client.controller import ClientController
    session = create_session(settings, args.debug)
    controller = ClientController(session)

    cmd = args.action
    if cmd == 'list':
        for client in controller.list_clients():
            print(client)
    else:
        try:
            if cmd == 'create':
                controller.create_client(args.name, args.appId, args.facets)
                print('Created client: %s' % args.name)
            elif cmd == 'show':
                client = controller.get_client(args.name)
                print('Client: %s' % client.name)
                print('AppID: %s' % client.app_id)
                print('FacetIDs:')
                for facet in client.valid_facets:
                    print('  %s' % facet)
            elif cmd == 'update':
                controller.update_client(args.name, args.appId, args.facets)
                print('Updated client: %s' % args.name)
            elif cmd == 'delete':
                controller.delete_client(args.name)
                print('Deleted client: %s' % args.name)
            session.commit()
        except Exception as e:
            print(e)
            if args.debug:
                raise e
            sys.exit(1)


def handle_run(settings, args):
    from u2fval.core.api import create_application
    from u2fval.client.pathinfo_auth import client_from_pathinfo
    from u2fval.client.controller import ClientController
    from wsgiref.simple_server import make_server

    extra_environ = {}
    if args.client:
        # Ensure the existance of the client.
        session = create_session(settings, args.debug)
        controller = ClientController(session)
        controller.get_client(args.client)
        session.close()
        print("Running in single-client mode for client: '%s'" % args.client)
        extra_environ['REMOTE_USER'] = args.client
        application = create_application(settings)
    else:
        application = client_from_pathinfo(create_application(settings))
    httpd = make_server(args.interface, args.port, application)
    httpd.base_environ.update(extra_environ)
    print("Starting server on http://%s:%d..." % (args.interface, args.port))
    httpd.serve_forever()


def handle_db(settings, args):
    from u2fval.model import Base
    from sqlalchemy import create_engine

    engine = create_engine(settings['db'], echo=args.debug)
    if args.action == 'init':
        Base.metadata.create_all(engine)
        print("Database intialized!")
    elif args.action == 'upgrade':
        try:
            from alembic import config, command
        except ImportError:
            print("Upgrading the database requires alembic")
            return
        conf = config.Config('alembic.ini')
        with engine.begin() as connection:
            conf.attributes['connection'] = connection
            command.upgrade(conf, 'head')
            print("Database upgraded to latest version")


def handle_args(settings, args):
    cmd = args.command
    if cmd == 'client':
        handle_client(settings, args)
    elif cmd == 'run':
        handle_run(settings, args)
    elif cmd == 'db':
        handle_db(settings, args)


def main():
    args = arg_parser().parse_args()

    if args.config:
        os.environ['U2FVAL_SETTINGS'] = args.config
    from u2fval.config import settings

    handle_args(settings, args)


if __name__ == '__main__':
    main()



# Copyright (c) 2014 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from sqlalchemy import (Column, Integer, String, Text, ForeignKey, Sequence,
                        Boolean, DateTime, BigInteger, UniqueConstraint)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relationship, backref, object_session
from sqlalchemy.orm.collections import attribute_mapped_collection

from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.serialization import Encoding
from uuid import uuid4
from datetime import datetime
from hashlib import sha1
from binascii import b2a_hex, b2a_base64, a2b_base64
import json


Base = declarative_base()


class Client(Base):
    __tablename__ = 'clients'

    id = Column(Integer, Sequence('client_id_seq'), primary_key=True)
    name = Column(String(32), nullable=False, unique=True)
    app_id = Column(String(256), nullable=False)
    _valid_facets = Column('valid_facets', Text(), default='[]')

    def __init__(self, name, app_id, facets):
        self.name = name
        self.app_id = app_id
        self.valid_facets = facets

    @hybrid_property
    def valid_facets(self):
        return json.loads(self._valid_facets)

    @valid_facets.setter
    def valid_facets(self, facets):
        if not isinstance(facets, list):
            raise TypeError('facets must be a list')
        self._valid_facets = json.dumps(facets)


class User(Base):
    __tablename__ = 'users'
    __table_args__ = (UniqueConstraint('client_id', 'name',
                                       name='_client_user_uc'),)

    id = Column(Integer, Sequence('user_id_seq'), primary_key=True)
    name = Column(String(40), nullable=False)
    client_id = Column(Integer, ForeignKey('clients.id'))
    client = relationship(Client, backref=backref('users'))
    devices = relationship(
        'Device',
        backref='user',
        order_by='Device.handle',
        collection_class=attribute_mapped_collection('handle'),
        cascade='all, delete-orphan'
    )
    transactions = relationship(
        'Transaction',
        backref='user',
        order_by='Transaction.created_at.desc()',
        lazy='dynamic',
        cascade='all, delete-orphan')

    def __init__(self, name):
        if len(name) > 40:
            self.name = sha1(name).hexdigest()
        else:
            self.name = name

    def add_device(self, bind_data, cert, transports=0):
        certificate = object_session(self).query(Certificate) \
            .filter(Certificate.fingerprint == b2a_hex(cert.fingerprint(
                hashes.SHA1()))) \
            .first()
        if certificate is None:
            certificate = Certificate(cert)
        return Device(self, bind_data, certificate, transports)


class Certificate(Base):
    __tablename__ = 'certificates'

    id = Column(Integer, Sequence('certificate_id_seq'), primary_key=True)
    fingerprint = Column(String(32), nullable=False, unique=True)
    _der = Column('der', Text(), nullable=False)

    @hybrid_property
    def der(self):
        return a2b_base64(self._der)

    @der.setter
    def der(self, der):
        self._der = b2a_base64(der)

    def __init__(self, cert):
        self.fingerprint = b2a_hex(cert.fingerprint(hashes.SHA1()))
        self.der = cert.public_bytes(Encoding.DER)


class Device(Base):
    __tablename__ = 'devices'

    id = Column(Integer, Sequence('device_id_seq'), primary_key=True)
    handle = Column(String(32), nullable=False, unique=True)
    user_id = Column(Integer, ForeignKey('users.id'))
    bind_data = Column(Text())
    certificate_id = Column(Integer, ForeignKey('certificates.id'))
    certificate = relationship('Certificate')
    compromised = Column(Boolean, default=False)
    created_at = Column(DateTime, default=datetime.utcnow)
    authenticated_at = Column(DateTime)
    counter = Column(BigInteger)
    transports = Column(BigInteger)
    _properties = relationship(
        'Property',
        backref='device',
        order_by='Property.key',
        collection_class=attribute_mapped_collection('key'),
        cascade='all, delete-orphan'
    )
    properties = association_proxy(
        '_properties',
        'value',
        creator=lambda k, v: Property(k, v)
    )

    def __init__(self, user, bind_data, certificate, transports=0):
        self.handle = uuid4().hex
        self.bind_data = bind_data
        self.user = user
        self.certificate = certificate
        self.transports = transports

    def get_descriptor(self, metadata=None):
        authenticated = self.authenticated_at
        if authenticated is not None:
            authenticated = authenticated.isoformat() + 'Z'

        data = {
            'handle': self.handle,
            'transports': self.transports,
            'compromised': self.compromised,
            'created': self.created_at.isoformat() + 'Z',
            'lastUsed': authenticated,
            'properties': dict(self.properties)
        }

        if metadata is not None:
            data['metadata'] = metadata

        return data


class Property(Base):
    __tablename__ = 'properties'

    id = Column(Integer, Sequence('property_id_seq'), primary_key=True)
    key = Column(String(32))
    value = Column(Text())
    device_id = Column(Integer, ForeignKey('devices.id'))

    def __init__(self, key, value):
        self.key = key
        self.value = value


# Used for storing transactions in the DB instead of memcached
# See transactiondb.py for more details.
class Transaction(Base):
    __tablename__ = 'transactions'

    id = Column(Integer, Sequence('transaction_id_seq'), primary_key=True)
    user_id = Column(Integer, ForeignKey('users.id'))
    transaction_id = Column(String(64), nullable=False, unique=True)
    _data = Column(Text())
    created_at = Column(DateTime, default=datetime.utcnow)

    def __init__(self, transaction_id, data):
        self.transaction_id = transaction_id
        self.data = data

    @hybrid_property
    def data(self):
        return json.loads(self._data)

    @data.setter
    def data(self, value):
        self._data = json.dumps(value)



# Database configuration string
DATABASE_CONFIGURATION = 'sqlite:///:memory:'

# If True, use memcached for storing registration and authentication requests
# in progress, instead of persisting them to the database.
USE_MEMCACHED = False

# If memcached is enabled, use these servers.
MEMCACHED_SERVERS = ['127.0.0.1:11211']

# Add files containing trusted metadata JSON to the directory below.
METADATA = '/etc/yubico/u2fval/metadata/'

# Allow the use of untrusted (for which attestation cannot be verified using
# the available trusted metadata) U2F devices.
ALLOW_UNTRUSTED = False



# Copyright (c) 2014 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import sys
import imp
import errno
import os
from u2fval import default_settings
import logging
import logging.config

__all__ = [
    'settings'
]

SETTINGS_FILE = os.getenv('U2FVAL_SETTINGS', os.path.join(
                          '/etc/yubico/u2fval/u2fval.conf'))
LOG_CONFIG_FILE = os.path.join(os.path.dirname(os.path.abspath(SETTINGS_FILE)),
                               'logging.conf')

VALUES = {
    'DATABASE_CONFIGURATION': 'db',
    'USE_MEMCACHED': 'mc',
    'MEMCACHED_SERVERS': 'mc_hosts',
    'METADATA': 'metadata',
    'ALLOW_UNTRUSTED': 'allow_untrusted'
}


def parse(conf, settings={}):
    for confkey, settingskey in VALUES.items():
        try:
            settings[settingskey] = conf.__getattribute__(confkey)
        except AttributeError:
            pass
    return settings


settings = parse(default_settings)

dont_write_bytecode = sys.dont_write_bytecode
try:
    sys.dont_write_bytecode = True
    user_settings = imp.load_source('user_settings', SETTINGS_FILE)
    settings = parse(user_settings, settings)
except IOError as e:
    if e.errno not in [errno.ENOENT, errno.EACCES]:
        raise e
finally:
    sys.dont_write_bytecode = dont_write_bytecode

# Set up logging
try:
    logging.config.fileConfig(LOG_CONFIG_FILE)
except:
    logging.basicConfig(level=logging.INFO)
    log = logging.getLogger(__name__)
    log.warning("Unable to configure logging. Logging to console.")



from webtest import TestApp
from u2fval.model import Base
from u2fval.core.api import create_application
from u2fval.client.controller import ClientController
from u2fval.core import exc
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from .soft_u2f_v2 import SoftU2FDevice
import unittest
import os


class RestApiTest(unittest.TestCase):

    def setUp(self):
        os.environ['U2FVAL_SETTINGS'] = '/dev/null'
        from u2fval.config import settings
        settings['allow_untrusted'] = True

        engine = create_engine(settings['db'])
        Base.metadata.create_all(engine)
        Session = sessionmaker(bind=engine)
        session = Session()

        self.client_controller = ClientController(session)
        self.client_controller.create_client('fooclient',
                                             'https://example.com',
                                             ['https://example.com'])

        self.app = TestApp(create_application(settings, session))

    def test_call_without_client(self):
        err = self.app.get('/', status=400)
        assert err.json['errorCode'] == exc.BadInputException.code

    def test_call_with_invalid_client(self):
        err = self.app.get('/', status=400,
                           extra_environ={'REMOTE_USER': 'invalid'})
        assert err.json['errorCode'] == exc.BadInputException.code

    def test_get_trusted_facets(self):
        resp = self.app.get('/', extra_environ={'REMOTE_USER': 'fooclient'})
        assert 'https://example.com' in resp.json['trustedFacets'][0]['ids']

    def test_list_empty_devices(self):
        resp = self.app.get('/foouser',
                            extra_environ={'REMOTE_USER': 'fooclient'})
        assert resp.json == []

    def test_begin_auth_without_devices(self):
        err = self.app.get('/foouser/authenticate', status=400,
                           extra_environ={'REMOTE_USER': 'fooclient'})
        assert err.json['errorCode'] == exc.NoEligibleDevicesException.code

    def test_register(self):
        device = SoftU2FDevice()
        self.do_register(device, {'foo': 'bar'})

    def test_authenticate(self):
        device = SoftU2FDevice()
        self.do_register(device, {'foo': 'bar', 'baz': 'one'})
        descriptor = self.do_authenticate(device, {'baz': 'two'})
        assert descriptor['properties'] == {
            'foo': 'bar',
            'baz': 'two'
        }

    def test_get_properties(self):
        device = SoftU2FDevice()
        descriptor = self.do_register(device, {'foo': 'bar', 'baz': 'foo'})
        descriptor2 = self.app.get('/foouser/' + descriptor['handle'],
                                   extra_environ={'REMOTE_USER': 'fooclient'})
        assert descriptor2.json['properties'] == {'foo': 'bar', 'baz': 'foo'}

    def test_get_devices(self):
        self.do_register(SoftU2FDevice())
        self.do_register(SoftU2FDevice())
        self.do_register(SoftU2FDevice())

        resp = self.app.get('/foouser',
                            extra_environ={'REMOTE_USER': 'fooclient'})
        descriptors = resp.json
        assert len(descriptors) == 3

    def test_delete_user(self):
        self.do_register(SoftU2FDevice())
        self.do_register(SoftU2FDevice())
        self.do_register(SoftU2FDevice())
        self.app.delete('/foouser',
                        extra_environ={'REMOTE_USER': 'fooclient'})
        resp = self.app.get('/foouser',
                            extra_environ={'REMOTE_USER': 'fooclient'})
        assert resp.json == []

    def test_delete_devices(self):
        d1 = self.do_register(SoftU2FDevice())
        d2 = self.do_register(SoftU2FDevice())
        d3 = self.do_register(SoftU2FDevice())

        self.app.delete('/foouser/' + d2['handle'],
                        extra_environ={'REMOTE_USER': 'fooclient'})
        resp = self.app.get('/foouser',
                            extra_environ={'REMOTE_USER': 'fooclient'})
        assert len(resp.json) == 2
        self.app.delete('/foouser/' + d1['handle'],
                        extra_environ={'REMOTE_USER': 'fooclient'})
        resp = self.app.get('/foouser',
                            extra_environ={'REMOTE_USER': 'fooclient'})
        assert resp.json == [d3]
        self.app.delete('/foouser/' + d3['handle'],
                        extra_environ={'REMOTE_USER': 'fooclient'})
        resp = self.app.get('/foouser',
                            extra_environ={'REMOTE_USER': 'fooclient'})
        assert resp.json == []

    def do_register(self, device, properties=None):
        reg_req = self.app.get('/foouser/register',
                               extra_environ={'REMOTE_USER': 'fooclient'})
        assert len(reg_req.json['authenticateRequests']) == \
            len(reg_req.json['authenticateDescriptors'])

        reg_resp = device.register(reg_req.json['registerRequests'][0],
                                   'https://example.com')

        if properties is None:
            properties = {}
        descriptor = self.app.post_json('/foouser/register', {
            'registerResponse': reg_resp.json,
            'properties': properties
        }, extra_environ={'REMOTE_USER': 'fooclient'})
        assert descriptor.json['properties'] == properties
        return descriptor.json

    def do_authenticate(self, device, properties=None):
        aut_req = self.app.get('/foouser/authenticate',
                               extra_environ={'REMOTE_USER': 'fooclient'})
        aut_resp = device.getAssertion(aut_req.json['authenticateRequests'][0],
                                       'https://example.com')
        if properties is None:
            properties = {}
        return self.app.post_json('/foouser/authenticate', {
            'authenticateResponse': aut_resp.json,
            'properties': properties
        }, extra_environ={'REMOTE_USER': 'fooclient'}).json



# Copyright (c) 2014 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


from u2fval.model import User, Transaction
from u2fval.core.exc import BadInputException
from datetime import datetime, timedelta
from binascii import b2a_hex


class DBStore(object):

    def __init__(self, session, max_transactions=5, ttl=300):
        self._session = session
        self._max_transactions = max_transactions
        self._ttl = ttl

    def _delete_expired(self):
        expiration = datetime.utcnow() - timedelta(seconds=self._ttl)
        self._session.query(Transaction) \
            .filter(Transaction.created_at < expiration).delete()

    def store(self, client_id, user_id, transaction_id, data):
        transaction_id = b2a_hex(transaction_id)
        user = self._session.query(User) \
            .filter(User.client_id == client_id) \
            .filter(User.name == user_id).first()
        if user is None:
            user = User(user_id)
            user.client_id = client_id
            self._session.add(user)
        else:
            self._delete_expired()
            # Delete oldest transactions until we have room for one more.
            for transaction in user.transactions \
                    .offset(self._max_transactions - 1).all():
                self._session.delete(transaction)
        user.transactions.append(Transaction(transaction_id, data))

    def retrieve(self, client_id, user_id, transaction_id):
        transaction_id = b2a_hex(transaction_id)
        self._delete_expired()
        transaction = self._session.query(Transaction) \
            .filter(Transaction.transaction_id == transaction_id).first()
        if transaction is None:
            raise BadInputException('Invalid transaction')
        if transaction.user.name != user_id or \
                transaction.user.client_id != client_id:
            raise BadInputException('Transaction not valid for user_id: %s'
                                    % user_id)
        self._session.delete(transaction)
        return transaction.data



# Copyright (c) 2014 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import memcache
from binascii import b2a_hex

__all__ = ['MemcachedStore']


class MemcachedStore(object):
    def __init__(self, hosts, max_transactions=5, ttl=300):
        self._mc = memcache.Client(hosts)
        self._max_transactions = max_transactions
        self._ttl = ttl

    def store(self, client_id, user_id, transaction_id, data):
        mc_key = '%s/%s' % (client_id, user_id)
        transaction_id = b2a_hex(transaction_id)
        keys = self._mc.get(mc_key) or []
        if len(keys) + 1 >= self._max_transactions:
            self._mc.delete('%s_%s' % (mc_key, keys.pop(0)))
        keys.append(transaction_id)
        t_key = '%s_%s' % (mc_key, transaction_id)
        self._mc.set_multi({
            mc_key: keys,
            t_key: data
        }, self._ttl)

    def retrieve(self, client_id, user_id, transaction_id):
        mc_key = '%s/%s' % (client_id, user_id)
        transaction_id = b2a_hex(transaction_id)
        t_key = '%s_%s' % (mc_key, transaction_id)
        data = self._mc.get_multi([mc_key, t_key])
        keys = data.get(mc_key)
        value = data[t_key]
        if keys:
            keys.remove(transaction_id)
            self._mc.set(mc_key, keys, self._ttl)
        if value:
            self._mc.delete(t_key)
        return value



# Copyright (c) 2014 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from u2fval.model import Device
from u2fval.core.controller import U2FController
from u2fval.core.jsobjects import (
    RegisterRequestData, RegisterResponseData, AuthenticateRequestData,
    AuthenticateResponseData)
from u2fval.core.exc import U2fException, BadInputException
from webob.dec import wsgify
from webob import exc, Response
from cachetools import LRUCache
from cachetools.func import lru_cache
from cryptography import x509
from cryptography.hazmat.backends import default_backend
import json
import logging


log = logging.getLogger(__name__)
__all__ = ['create_application']


def u2f_error(e):
    server_e = exc.HTTPBadRequest()
    server_e.body = e.json.encode('utf8')
    server_e.content_type = 'application/json'
    return server_e


class U2FServerApplication(object):

    def __init__(self, session, memstore, metadata, allow_untrusted=False):
        self._session = session
        self._memstore = memstore
        self._metadata = metadata
        self._require_trusted = not allow_untrusted

    @wsgify
    def __call__(self, request):
        client_name = request.environ.get('REMOTE_USER')
        if not client_name:
            raise u2f_error(BadInputException('Client not specified'))
        try:
            resp = self.client(request, client_name)
            if not isinstance(resp, Response):
                resp = Response(json.dumps(resp),
                                charset='utf-8',
                                content_type='application/json')
            return resp
        except Exception as e:
            self._session.rollback()
            if isinstance(e, U2fException):
                e = u2f_error(e)
            elif isinstance(e, exc.HTTPException):
                pass
            else:
                log.exception('Server error')
                e = exc.HTTPServerError(e.args[0])
            raise e
        finally:
            self._session.commit()

    @lru_cache(maxsize=16)
    def _get_controller(self, client_name):
        try:
            return U2FController(self._session, self._memstore, client_name,
                                 self._metadata, self._require_trusted)
        except KeyError as e:
            raise BadInputException(e.args[0])

    def client(self, request, client_name):
        user_id = request.path_info_pop()
        controller = self._get_controller(client_name)
        if not user_id:
            if request.method == 'GET':
                return controller.get_trusted_facets()
            else:
                raise exc.HTTPMethodNotAllowed
        return self.user(request, controller, user_id.encode('utf-8'))

    def user(self, request, controller, user_id):
        if request.path_info_peek():
            page = request.path_info_pop()
            if page == 'register':
                return self.register(request, controller, user_id)
            elif page == 'authenticate':
                return self.authenticate(request, controller, user_id)
            else:
                return self.device(request, controller, user_id, page)

        if request.method == 'GET':
            return controller.get_descriptors(user_id)
        elif request.method == 'DELETE':
            controller.delete_user(user_id)
            return exc.HTTPNoContent()
        else:
            raise exc.HTTPMethodNotAllowed

    def register(self, request, controller, user_id):
        if request.method == 'GET':
            register_requests, sign_requests, descriptors = controller \
                .register_start(user_id)
            return RegisterRequestData(
                registerRequests=register_requests,
                authenticateRequests=sign_requests,
                authenticateDescriptors=descriptors
            )
        elif request.method == 'POST':
            data = RegisterResponseData(request.body)
            try:
                handle = controller.register_complete(user_id,
                                                      data.registerResponse)
            except KeyError:
                raise exc.HTTPBadRequest
            controller.set_props(handle, data.properties)

            return controller.get_descriptor(user_id, handle)
        else:
            raise exc.HTTPMethodNotAllowed

    def authenticate(self, request, controller, user_id):
        if request.method == 'GET':
            sign_requests, descriptors = controller.authenticate_start(user_id)
            return AuthenticateRequestData(
                authenticateRequests=sign_requests,
                authenticateDescriptors=descriptors
            )
        elif request.method == 'POST':
            data = AuthenticateResponseData(request.body)
            try:
                handle = controller.authenticate_complete(
                    user_id, data.authenticateResponse)
            except KeyError:
                raise BadInputException('Malformed request')
            except ValueError as e:
                log.exception('Error in authenticate')
                raise BadInputException(e.message)
            controller.set_props(handle, data.properties)

            return controller.get_descriptor(user_id, handle)
        else:
            raise exc.HTTPMethodNotAllowed

    def device(self, request, controller, user_id, handle):
        try:
            if request.method == 'GET':
                return controller.get_descriptor(user_id, handle)
            elif request.method == 'POST':
                props = json.loads(request.body)
                controller.set_props(handle, props)
                return controller.get_descriptor(user_id, handle)
            elif request.method == 'DELETE':
                controller.unregister(handle)
                return exc.HTTPNoContent()
            else:
                raise exc.HTTPMethodNotAllowed
        except ValueError as e:
            raise exc.HTTPNotFound(e.message)


class MetadataCache(object):

    def __init__(self, provider, maxsize=64):
        self._provider = provider
        self._cache = LRUCache(maxsize=maxsize)

    def get_attestation(self, device_or_cert):
        if isinstance(device_or_cert, Device):
            device = device_or_cert
            if device.certificate_id not in self._cache:
                cert = x509.load_der_x509_certificate(device.certificate.der,
                                                      default_backend())
                attestation = self._provider.get_attestation(cert)
                self._cache[device.certificate_id] = attestation
            return self._cache[device.certificate_id]
        else:
            return self._provider.get_attestation(device_or_cert)

    def get_metadata(self, device):
        attestation = self.get_attestation(device)
        if attestation:
            metadata = {}
            if attestation.vendor_info:
                metadata['vendor'] = attestation.vendor_info
            if attestation.device_info:
                metadata['device'] = attestation.device_info
            return metadata
        return None


def create_application(settings, session=None):
    if not session:
        from sqlalchemy import create_engine
        from sqlalchemy.orm import sessionmaker
        engine = create_engine(settings['db'], echo=False)

        Session = sessionmaker(bind=engine)
        session = Session()

    from u2flib_server.attestation import MetadataProvider, create_resolver
    import os
    data = settings['metadata']
    # If pointing to an empty or non-existant directory, set to None so that
    # built-in metadata is used.
    try:
        allowed_types = (basestring,)  # Python 2
    except NameError:
        allowed_types = (str, bytes)  # Python 3
    if isinstance(data, allowed_types) \
            and not os.path.isfile(data) \
            and (not os.path.isdir(data) or len(os.listdir(data)) == 0):
        data = None
    metadata = MetadataCache(MetadataProvider(create_resolver(data)))

    if settings['mc']:
        from u2fval.core.transactionmc import MemcachedStore
        memstore = MemcachedStore(settings['mc_hosts'])
    else:
        from u2fval.core.transactiondb import DBStore
        memstore = DBStore(session)

    return U2FServerApplication(session, memstore, metadata,
                                settings['allow_untrusted'])



# Copyright (c) 2014 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import print_function
from u2fval.yubicommon.setup import setup
import sys
import os
import glob


# Make sure the cont/u2fval.conf file exists and is a copy of
# u2fval/default_settings.py
if [_ for _ in ['install', 'sdist'] if _ in sys.argv[1:]]:
    print("copying default settings...")
    source = os.path.abspath('u2fval/default_settings.py')
    target = os.path.abspath('conf/u2fval.conf')
    with open(target, 'w') as target_f:
        with open(source, 'r') as source_f:
            target_f.write(source_f.read())
    os.chmod(target, 0o600)


def can_write_etc(path=os.path.join(os.path.sep, 'etc', 'yubico', 'u2fval')):
    if not os.path.isdir(path):
        return can_write_etc(os.path.dirname(path))
    return os.access(path, os.W_OK)


# Only write configuration files if we have the correct permissions.
data_files = [
    ('/etc/yubico/u2fval', ['conf/u2fval.conf', 'conf/logging.conf']),
    ('/etc/yubico/u2fval/metadata', glob.glob('conf/metadata/*.json'))
] if can_write_etc() else []

setup(
    name='u2fval',
    author='Dain Nilsson',
    author_email='dain@yubico.com',
    maintainer='Yubico Open Source Maintainers',
    maintainer_email='ossmaint@yubico.com',
    description='Standalone/WSGI U2F server implementing the U2FVAL protocol',
    url='https://github.com/Yubico/u2fval',
    license='BSD 2 clause',
    entry_points={
        'console_scripts': ['u2fval=u2fval.cli:main']
    },
    data_files=data_files,
    install_requires=['python-u2flib-server >= 4, <5', 'SQLAlchemy',
                      'WebOb', 'cachetools'],
    test_suite='test',
    tests_require=['WebTest'],
    extras_require={
        'u2fval:python_version=="2.6"': ['argparse'],
        'memcache': ['python-memcached'],
        'db_migration': ['alembic']
    },
    classifiers=[
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Internet',
        'Topic :: Security',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Application'
    ]
)



# Copyright (c) 2014 Yubico AB
# All rights reserved.
#
#   Redistribution and use in source and binary forms, with or
#   without modification, are permitted provided that the following
#   conditions are met:
#
#    1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


from u2fval.model import Client
from sqlalchemy.orm import exc
import re
import logging


__all__ = ['ClientController']
log = logging.getLogger(__name__)


NAME_PATTERN = re.compile(r'^[a-zA-Z0-9-_.]{3,}$')


def ensure_valid_name(name):
    if len(name) < 3:
        raise ValueError('Client names must be at least 3 characters')
    if not NAME_PATTERN.match(name):
        raise ValueError('Client names may only contain the characters a-z, '
                         'A-Z, 0-9, "." (period), "_" (underscore), and "-" '
                         '(dash)')


class ClientController(object):

    def __init__(self, session):
        self._session = session

    def get_client(self, name):
        try:
            return self._session.query(Client) \
                .filter(Client.name == name) \
                .one()
        except exc.NoResultFound:
            raise KeyError('No Client with name %s found' % name)

    def create_client(self, name, app_id, valid_facets):
        ensure_valid_name(name)

        try:
            self.get_client(name)
            raise ValueError('Client already exists: %s' % name)
        except KeyError:
            client = Client(name, app_id, valid_facets)
            self._session.add(client)
            log.info('Client created: %s', name)

    def update_client(self, name, app_id=None, valid_facets=None):
        client = self.get_client(name)
        if app_id is not None:
            client.app_id = app_id
        if valid_facets is not None:
            client.valid_facets = valid_facets

    def delete_client(self, name):
        client = self.get_client(name)
        self._session.delete(client)
        log.info('Client deleted: %s', name)

    def list_clients(self):
        return [c[0] for c in self._session.query(Client.name).all()]



"""Compatibility constants and helpers for Python 2.x and 3.x.
"""

import sys

# NB If this module grows to more than a handful of items it is probably
#    to bite the bullet and depend on the six package.
__all__ = [
    'string_types',
    'binary_type',
    'text_type',
    'int2byte',
    'byte2int'
]

# Needed for isinstance() checks
# Same behaviour as six.string_types https://pythonhosted.org/six/#constants
if sys.version_info < (3, 0):
    # Python 2.x
    _PY2 = True
    string_types = (basestring,)
    binary_type = str
    text_type = unicode
else:
    # Python 3.x
    _PY2 = False
    string_types = (str,)
    binary_type = bytes
    text_type = str

def int2byte(i):
    if _PY2:
        return chr(i)
    return bytes((i,))

def byte2int(i):
    if _PY2:
        return ord(i)
    return i
